// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// ndmrlat.cpp - NetworkDirect memory registration latency test
//

#include "ndcommon.h"


const SIZE_T x_MaxSize = (4 * 1024 * 1024);
const SIZE_T x_Iterations = 10000;


void ShowUsage()
{
    printf( "ndmrlat <ip> [l<log>]\n"
        "\t<ip> - IPv4 Address\n"
        "\tl - log output to a file named <log>.\n");
}


void RunTest( INDAdapter* pAdapter, HANDLE hIocp, OVERLAPPED* pOv )
{
    char* buf = new (std::nothrow) char[x_MaxSize];
    if( buf == NULL )
    {
        printf( "Failed to allocate buffer.\n" );
        exit( __LINE__ );
    }

    ND_MR_HANDLE hMr;

    CpuMonitor cpu;

    Timer timer;

    double totalRegTime;
    double totalRegCallTime;
    double totalDeregTime;
    double totalDeregCallTime;

    for( SIZE_T szXfer = 1; szXfer <= x_MaxSize; szXfer <<= 1 )
    {
        totalRegTime = 0;
        totalRegCallTime = 0;
        totalDeregTime = 0;
        totalDeregCallTime = 0;

        cpu.Start();

        for( SIZE_T i = 0; i < x_Iterations; i++ )
        {
            //
            // Register
            //
            timer.Start();
            HRESULT hr = pAdapter->RegisterMemory( buf, szXfer, pOv, &hMr );
            timer.Split();

            if( FAILED( hr ) )
            {
                printf( "INDAdapter::RegisterMemory failed with %08x\n", hr );
                exit( __LINE__ );
            }

            if( pOv->hEvent == NULL )
            {
                DWORD BytesRet2;
                ULONG_PTR Key;
                OVERLAPPED* pOv2;
                BOOL fSuccess = GetQueuedCompletionStatus(
                    hIocp,
                    &BytesRet2,
                    &Key,
                    &pOv2,
                    INFINITE
                    );

                if( fSuccess == FALSE )
                {
                    printf( "GetQueuedCompletionStatus failed with %u\n", GetLastError() );
                    exit( __LINE__ );
                }
                if( pOv2 != pOv )
                {
                    printf(
                        "Unexpected overlapped %p returned but expected %p\n",
                        pOv2,
                        pOv
                        );
                    exit( __LINE__ );
                }
            }

            SIZE_T BytesRet;
            hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, pOv->hEvent != NULL );
            timer.End();

            if( FAILED( hr ) )
            {
                printf( "INDAdapter->GetOverlappedResult failed with %08x\n", hr );
                exit( __LINE__ );
            }

            totalRegCallTime += timer.ReportPreSplit();
            totalRegTime += timer.Report();

            //
            // Deregister
            //
            timer.Start();
            hr = pAdapter->DeregisterMemory( hMr, pOv );
            timer.Split();

            if( FAILED( hr ) )
            {
                printf( "INDAdapter::DeregisterMemory failed with %08x\n", hr );
                exit( __LINE__ );
            }

            if( pOv->hEvent == NULL )
            {
                DWORD BytesRet2;
                ULONG_PTR Key;
                OVERLAPPED* pOv2;
                BOOL fSuccess = GetQueuedCompletionStatus(
                    hIocp,
                    &BytesRet2,
                    &Key,
                    &pOv2,
                    INFINITE
                    );

                if( fSuccess == FALSE )
                {
                    printf( "GetQueuedCompletionStatus failed with %u\n", GetLastError() );
                    exit( __LINE__ );
                }
                if( pOv2 != pOv )
                {
                    printf(
                        "Unexpected overlapped %p returned but expected %p\n",
                        pOv2,
                        pOv
                        );
                    exit( __LINE__ );
                }
            }

            hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, pOv->hEvent != NULL );
            timer.End();

            if( FAILED( hr ) )
            {
                printf( "INDAdapter->GetOverlappedResult failed with %08x\n", hr );
                exit( __LINE__ );
            }

            totalDeregCallTime += timer.ReportPreSplit();
            totalDeregTime += timer.Report();
        }
        cpu.End();

        printf(
            "%9Id %9.2f %9.2f %9.2f %9.2f %7.2f\n",
            szXfer,
            totalRegCallTime / x_Iterations,
            totalRegTime / x_Iterations,
            totalDeregCallTime / x_Iterations,
            totalDeregTime / x_Iterations,
            cpu.Report()
            );
    }

    delete[] buf;
}


int __cdecl _tmain(int argc, TCHAR* argv[])
{
    struct sockaddr_in v4 = {0};

    WSADATA wsaData;
    int ret = ::WSAStartup( MAKEWORD( 2, 2 ), &wsaData );
    if( ret != 0 )
    {
        printf( "Failed to initialize Windows Sockets: %d\n", ret );
        exit( __LINE__ );
    }

    if( argc < 2 || argc > 3 )
    {
        ShowUsage();
        exit( __LINE__ );
    }

    INT len = sizeof( v4 );
    WSAStringToAddress(
        argv[1],
        AF_INET,
        NULL,
        reinterpret_cast<struct sockaddr*>( &v4 ),
        &len
        );

    if( argc == 3 )
    {
        if( argv[2][0] != 'l' &&
            argv[2][0] != 'L' )
        {
            ShowUsage();
            exit( __LINE__ );
        }

//
// Can't use freopen_s because it doesn't allow sharing.
// So supress the deprecated warning, because for our use
// it isn't deprecated.
//
#pragma warning( disable : 4996 )
        if( _tfreopen( &argv[2][1], _T("w"), stdout ) == NULL ||
            _tfreopen( &argv[2][1], _T("a+"), stderr ) == NULL )
#pragma warning( default : 4996 )
        {
            printf( "Could not open log file.\n" );
            exit( __LINE__ );
        }
    }

    if( v4.sin_addr.s_addr == 0 )
    {
        printf( "Bad address.\n");
        ShowUsage();
        exit( __LINE__ );
    }

    HRESULT hr = NdStartup();
    if( FAILED(hr) )
    {
        printf( "NdStartup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDAdapter* pAdapter;
    hr = NdOpenV1Adapter( (const struct sockaddr*)&v4, sizeof(v4), &pAdapter );
    if( FAILED( hr ) )
    {
        printf( "NdOpenV1Adapter failed with %08x\n", hr );
        exit( __LINE__ );
    }

    HANDLE hIocp = CreateIoCompletionPort( pAdapter->GetFileHandle(), NULL, 0, 0 );
    if( hIocp == NULL )
    {
        printf( "Failed to bind adapter to IOCP, error %u\n", GetLastError() );
        exit( __LINE__ );
    }

    printf(
        "Using %u processors. Sender Frequency is %I64d\n",
        CpuMonitor::CpuCount(),
        Timer::Frequency() );

    printf(
        "\n Event Driven:\n"
        "%29s %19s %7s\n"
        "%9s %9s %9s %9s %9s\n",
        "Register (usec)",
        "Deregister (usec)",
        "CPU",
        "Size",
        "call",
        "total",
        "call",
        "total"
        );

    OVERLAPPED Ov;

    //
    // Note that we run the test using GetOverlappedResult even though the adapter
    // is bound to an I/O completion port to make sure that the IOCP semantics are
    // working - that is, setting the lower bit of the hEvent member of the
    // OVERLAPPED structure prevents the completion from being reported to the IOCP.
    //
    Ov.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
    if( Ov.hEvent == NULL )
    {
        printf( "Create event failed with %u\n", GetLastError() );
        exit( __LINE__ );
    }

    //
    // Set the lower bit of the event handle so completions don't go to the IOCP.
    //
    Ov.hEvent = (HANDLE)(((SIZE_T)Ov.hEvent) | 0x1);

    RunTest( pAdapter, hIocp, &Ov );

    //
    // Now we run again, using the IOCP, to see if performance is any different.
    //
    CloseHandle( Ov.hEvent );
    Ov.hEvent = NULL;

    printf(
        "\n IOCP:\n"
        "%29s %19s %7s\n"
        "%9s %9s %9s %9s %9s\n",
        "Register (usec)",
        "Deregister (usec)",
        "CPU",
        "Size",
        "call",
        "total",
        "call",
        "total"
        );

    RunTest( pAdapter, hIocp, &Ov );

    pAdapter->Release();

    CloseHandle( hIocp );

    hr = NdCleanup();
    if( FAILED(hr) )
    {
        printf( "NdCleanup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    _fcloseall();
    WSACleanup();

    return 0;
}
