// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// ndmval.cpp - NetworkDirect memory validation test
//
// This test establishes a connection, sends 1024 bytes from the client, and receives with a 1023-byte on the server side
//
// Client operation:
//  1. Establish connection
//  2. Allocate 1024 bytes of buffer
//	3. Send 1024 bytes to server
//	4. Expect ND_SUCCESS on iWARP or ND_REMOTE_ERROR on IB
// Server operation:
//  1. Establish connection
//	2. Allocate 1023 bytes of buffer
//	3. Try to receive from client
//	4. Expect ND_BUFFER_OVERFLOW

#include <ntstatus.h>
#define WIN32_NO_STATUS
#include "ndcommon.h"
#include <logging.h>


HRESULT DisconnectConnector(INDConnector* pConnector,OVERLAPPED* pOv)
{
	HRESULT hr = pConnector->Disconnect( pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}

	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT( hr, L"INDEndpoint::Disconnect failed with %08x", __LINE__ );
	}
	pConnector->Release();
	return hr;
}

const USHORT x_DefaultPort = 54331;
const SIZE_T x_InsufficientXfer = (1023u);
const SIZE_T x_SufficientXfer= (1024u);

const LPWSTR TESTNAME = L"ndmval.exe";

enum NdResultIndex
{
	Recv,
	Send,
	Read,
	MwBind = Read,
	Write,
	MwInvalidate = Write,
	ResultCount
};


void ShowUsage()
{
	printf( "ndmval s|c <ip>:<port> [l<log>]\n"
		"\ts - start as server (listen on IP/Port)\n"
		"\tc - start as client (connect to server IP/Port)\n"
		"\t<ip> - IPv4 Address\n"
		"\t<port> - Port number\n"
		"\tl - log output to a file named <log>.\n");
}

void WaitForCompletion( INDCompletionQueue* pCq, OVERLAPPED* pOv, ND_RESULT** ppResult, int line )
{
	HRESULT hr = pCq->Notify( ND_CQ_NOTIFY_ANY, pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pCq->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}
	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDCompletionQueue::Notify failed with %08x", line );
	}

	pCq->GetResults( ppResult, 1 );
}

void Server(
	_In_ OVERLAPPED* pOv,
	_In_ const struct sockaddr_in& v4
	)
{
	INDAdapter *pAdapter;
	HRESULT hr = NdOpenV1Adapter( (const struct sockaddr*)&v4, sizeof(v4), &pAdapter );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"NdOpenV1Adapter failed with %08x", __LINE__ );
	}

	ND_ADAPTER_INFO Info;
	SIZE_T Len = sizeof(Info);
	hr = pAdapter->Query( 1, &Info, &Len );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::Query failed with %08x", __LINE__ );
	}

	//
	// Allocate and register the data buffer.
	//
	char* pBuf = (char*)HeapAlloc( GetProcessHeap(), 0, x_InsufficientXfer );
	if( !pBuf ) 
	{
		LOG_FAILURE_AND_EXIT( L"Failed to allocate data buffer.", __LINE__ );
	}

	ND_MR_HANDLE hMr;
	hr = pAdapter->RegisterMemory( pBuf, x_InsufficientXfer, pOv, &hMr );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}

	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::RegisterMemory failed with %08x", __LINE__ );
	}

	//
	// Listen and get incoming connection request.
	//
	INDListen *pListen;
	//first variable 0 indicates no limit for the backlog
	hr = pAdapter->Listen( 0, IPPROTO_TCP, v4.sin_port, NULL, &pListen );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::Listen failed with %08x", __LINE__ );
	}

	INDConnector* pConnector;
	hr = pAdapter->CreateConnector( &pConnector );
	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::CreateConnector failed with %08x", __LINE__ );
	}

	printf( "Waiting for connection.\n" );
	hr = pListen->GetConnectionRequest( pConnector, pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pListen->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}

	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDListen::GetConnectionRequest failed with %08x", __LINE__ );
	}

	SIZE_T Ird =1;
	SIZE_T Ord =1;
	
	//
	// Create a CQ.
	//
	INDCompletionQueue *pCq;
	hr = pAdapter->CreateCompletionQueue( 2, &pCq );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::CreateCompletionQueue failed with %08x", __LINE__ );
	}

	INDEndpoint* pEndpoint;
	hr = pConnector->CreateEndpoint(
		pCq,
		pCq,
		1,
		1,
		1,
		1,
		Ird,
		Ord,
		NULL,
		&pEndpoint
		);
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDConnector::CreateEndpoint failed with %08x", __LINE__ );
	}

	ND_RESULT* Results = new (std::nothrow) ND_RESULT[ResultCount];
	if( Results == NULL )
	{
		LOG_FAILURE_AND_EXIT( L"Failed to allocate ND_RESULT array.", __LINE__ );
	}

	//
	// Pre-post receive request.
	//
	// note here that pBuf has 1023 bytes, but were given a length of 1024
	ND_SGE Sge;
	Sge.pAddr = pBuf;
	Sge.Length = x_InsufficientXfer;
	Sge.hMr = hMr;
	hr = pEndpoint->Receive( &Results[Recv], &Sge, 1 );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDEndpoint::Receive failed with %08x", __LINE__ );
	}

	
	hr = pConnector->Accept(
		pEndpoint,
		NULL,
		0,
		pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T bytesRet;
		hr = pConnector->GetOverlappedResult( pOv, &bytesRet, TRUE );
	}

	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDConnector::Accept failed with %08x", __LINE__ );
	}

	printf( "Connected.\n" );

	//
	// Test starts
	//
	ND_RESULT* pResult;

	
	printf( "Read data from client with insufficient buffer\n" );

	//
	// Wait for receive completion.
	//
	WaitForCompletion( pCq, pOv, &pResult, __LINE__ );
	if( pResult != &Results[Recv] )
	{
		LOG_FAILURE_AND_EXIT( L"Unexpected completion - expected Recv.", __LINE__ );
	}

	if( pResult->Status != ND_DATA_OVERRUN )
	{
		if( pResult->Status == ND_SUCCESS)
        {
			LOG_FAILURE_AND_EXIT(
                L"INDEndpoint::Recv expected to fail with ND_BUFFER_OVERFLOW, but got ND_SUCCESS",
                __LINE__
                );
		}
        //
		// If we get the wrong error code, treat this as a warning
        //
		LOG_FAILURE_HRESULT(
            pResult->Status,
            L"Warning: INDEndpoint::Recv expected to fail with ND_DATA_OVERRUN, but got %08x",
            __LINE__
            );
	}

	//
    // Send 0 byte message to indicate test 1 complete.
    //
   /* hr = pEndpoint->Send( &Results[Send], NULL, 0, 0 );
    if( FAILED( hr ) )
    {
        LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDEndpoint::Send failed with %08x", __LINE__ );
    }*/

    //
    // Wait for send completion.
    //
    /*WaitForCompletion( pCq, pOv, &pResult, __LINE__ );
    if( pResult != &Results[Send] )
    {
        LOG_FAILURE_AND_EXIT( L"Unexpected completion - expected Send.", __LINE__ );
    }

    if( pResult->Status != ND_SUCCESS )
    {
        LOG_FAILURE_HRESULT_AND_EXIT( pResult->Status, L"INDEndpoint::Send failed with %08x", __LINE__ );
    }

    if( pResult->BytesTransferred != 0 )
    {
        LOG_FAILURE_HRESULT_AND_EXIT(
            (LONG)pResult->BytesTransferred,
            L"Sent %d bytes, expected 0",
            __LINE__
            );
    }*/
   
	printf( "Test complete\n" );
	//
	// Test is complete.  Note that during teardown we don't treat
	// errors as test failures, but we still report them.
	//
	
    hr=DisconnectConnector(pConnector,pOv);
	pEndpoint->Release();
	pCq->Release();
	pListen->Release();

	hr = pAdapter->DeregisterMemory( hMr, pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}
	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT( hr, L"INDAdapter::DeregisterMemory failed with %08x", __LINE__ );
	}
	pAdapter->Release();

	delete[] Results;
	HeapFree( GetProcessHeap(), 0, pBuf );
}


void Client(
	_In_ OVERLAPPED* pOv,
	_In_ const struct sockaddr_in& v4
	)
{
	struct sockaddr_in localAddr;

	SIZE_T Len = sizeof(localAddr);
	HRESULT hr = NdResolveAddress(
		(const struct sockaddr*)&v4,
		sizeof(v4),
		(struct sockaddr*)&localAddr,
		&Len );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"NdResolveAddress failed with %08x", __LINE__ );
	}

	INDAdapter* pAdapter;
	hr = NdOpenV1Adapter(
		(const struct sockaddr*)&localAddr,
		sizeof(localAddr),
		&pAdapter );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"NdOpenV1Adapter failed with %08x", __LINE__ );
	}

	ND_ADAPTER_INFO Info;
	Len = sizeof(Info);
	hr = pAdapter->Query( 1, &Info, &Len );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::Query failed with %08x", __LINE__ );
	}

	//
	// Allocate and register the data buffer.
	//
	char* pBuf = (char*)HeapAlloc( GetProcessHeap(), 0, x_SufficientXfer );
	if( !pBuf )
	{
		LOG_FAILURE_AND_EXIT( L"Failed to allocate data buffer.", __LINE__ );
	}

	ND_MR_HANDLE hMr;
	hr = pAdapter->RegisterMemory( pBuf, x_SufficientXfer, pOv, &hMr );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::RegisterMemory failed with %08x",__LINE__ );
	}

	//
	// Create a CQ.
	//
	INDCompletionQueue* pCq;
	hr = pAdapter->CreateCompletionQueue( 2, &pCq );
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::CreateCompletionQueue failed with %08x", __LINE__ );
	}

	INDConnector* pConnector;
	hr = pAdapter->CreateConnector( &pConnector );
	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDAdapter::CreateConnector failed with %08x", __LINE__ );
	}

	INDEndpoint* pEndpoint;
	hr = pConnector->CreateEndpoint(
		pCq,
		pCq,
		1,
		1,
		1,
		1,
		1,
		1,
		NULL,
		&pEndpoint
		);
	if( FAILED( hr ) ) 
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDConnector::CreateEndpoint failed with %08x", __LINE__ );
	}

	ND_RESULT* Results = new (std::nothrow) ND_RESULT[ResultCount];
	if( Results == NULL )
	{
		LOG_FAILURE_AND_EXIT( L"Failed to allocate ND_RESULT array.", __LINE__ );
	}

	//
	// Pre-post receive request for detecting test completion.
	//
	hr = pEndpoint->Receive( &Results[Recv], NULL, 0 );
	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDEndpoint::Receive failed with %08x", __LINE__ );
	}

	//
	// Connect to the server.
	//
	printf( "Connecting...\r" );
	
	hr = pConnector->Connect(
		pEndpoint,
		(const struct sockaddr*)&v4,
		sizeof(v4),
		IPPROTO_TCP,
		0,
		NULL,
		0,
		pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}
	if( FAILED( hr ) || hr == ND_TIMEOUT )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDEndpoint::Connect failed with %08x", __LINE__ );
	}

	
	//
	// Complete the connection - this transitions the endpoint so it can send.
	//
	hr = pConnector->CompleteConnect( pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}
	if( FAILED( hr ) || hr == ND_TIMEOUT )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDEndpoint::CompleteConnect failed with %08x", __LINE__ );
	}

	printf( "Connected.     \n" );

	ND_RESULT* pResult;

	ND_SGE Sgl;
	Sgl.pAddr = pBuf;
	Sgl.Length = x_SufficientXfer;
	Sgl.hMr = hMr;
  
	hr = pEndpoint->Send( &Results[Send], &Sgl, 1, 0 );
	if( FAILED( hr ) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"INDEndpoint::Send failed with %08x", __LINE__ );
	}

	//
	// Wait for send completion.
	//
	WaitForCompletion( pCq, pOv, &pResult, __LINE__ );
	if( pResult != &Results[Send] )
	{
		LOG_FAILURE_AND_EXIT( L"Unexpected completion - expected Send.", __LINE__ );
	}

    //
	// We expect status to be ND_SUCCESS (for iWARP) or ND_REMOTE_ERROR(for IB)
    //
#pragma prefast(suppress:6221, "Implicit cast from NTSTATUS to HRESULT OK.")
	if( pResult->Status != ND_SUCCESS && pResult->Status != STATUS_DATA_ERROR)
	{
		LOG_FAILURE_HRESULT_AND_EXIT( pResult->Status, L"INDEndpoint::Send expecting ND_SUCCESS or ND_REMOTE_ERROR because receiver doesn't have enough buffer, but got errors %08x", __LINE__ );
	}

	//
	// Wait for the receive.
	//
	WaitForCompletion( pCq, pOv, &pResult, __LINE__ );
	if( pResult != &Results[Recv] )
	{
		LOG_FAILURE_AND_EXIT( L"Unexpected completion - expected Recv.", __LINE__ );
	}

	if( pResult->Status != ND_CANCELED )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( pResult->Status, L"INDEndpoint::Recv should fail with ND_CANCELED, but got %08x", __LINE__ );
	}

	printf( "Test complete\n" );
	//
	// Test is complete.  Note that during teardown we don't treat
	// errors as test failures, but we still report them.
	//
	DisconnectConnector(pConnector,pOv);
	pEndpoint->Release();
	pCq->Release();

	hr = pAdapter->DeregisterMemory( hMr, pOv );
	if( hr == ND_PENDING )
	{
		SIZE_T BytesRet;
		hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
	}
	if( FAILED( hr ) ) 
		LOG_FAILURE_HRESULT( hr, L"INDAdapter::DeregisterMemory failed with %08x", __LINE__ );

	pAdapter->Release();

	delete[] Results;
	HeapFree( GetProcessHeap(), 0, pBuf );
}

int __cdecl _tmain(int argc, TCHAR* argv[])
{

	bool isServer = false;
	bool isClient = false;
	struct sockaddr_in v4Server = {0};

	WSADATA wsaData;
	int ret = ::WSAStartup( MAKEWORD( 2, 2 ), &wsaData );
	if( ret != 0 )
	{
		printf( "Failed to initialize Windows Sockets: %d\n", ret );
		exit( __LINE__ );
	}

	INIT_LOG( TESTNAME );
	const TCHAR* pArg;
	int len;
	for( int i = 1; i < argc; i++ )
	{
		pArg = argv[i];

		// Skip leading dashes/slashes
		while( *pArg == '-' || *pArg == '/' )
		{
			pArg++;
		}

		switch( *pArg )
		{
		case _T('s'):
		case _T('S'):
			isServer = true;
			if( ++i == argc )
			{
				break;
			}
			len = sizeof( v4Server );
			WSAStringToAddress(
				argv[i],
				AF_INET,
				NULL,
				reinterpret_cast<struct sockaddr*>( &v4Server ),
				&len
				);
			break;
		case _T('c'):
		case _T('C'):
			isClient = true;
			if( ++i == argc )
			{
				break;
			}
			len = sizeof( v4Server );
			WSAStringToAddress(
				argv[i],
				AF_INET,
				NULL,
				reinterpret_cast<struct sockaddr*>( &v4Server ),
				&len
				);
			break;
		case _T('l'):
		case _T('L'):
//
// Can't use freopen_s because it doesn't allow sharing.
// So supress the deprecated warning, because for our use
// it isn't deprecated.
//
#pragma warning( disable : 4996 )
			if( _tfreopen( ++pArg, _T("w"), stdout ) == NULL ||
				_tfreopen( pArg, _T("a+"), stderr ) == NULL )
#pragma warning( default : 4996 )
			{
				printf( "Could not open log file.\n" );
				exit( __LINE__ );
			}
			break;
		default:
			_tprintf( _T("Unknown parameter %s\n"), pArg );
			ShowUsage();
			exit( __LINE__ );
		}
	}

	if( (isClient && isServer) || (!isClient && !isServer) )
	{
		printf( "Exactly one of client (c or "
			"server (s) must be specified.\n" );
		ShowUsage();
		exit( __LINE__ );
	}

	if( v4Server.sin_addr.s_addr == 0 )
	{
		printf( "Bad address.\n\n");
		ShowUsage();
		exit( __LINE__ );
	}

	if( v4Server.sin_port == 0 )
	{
		v4Server.sin_port = htons( x_DefaultPort );
	}

	OVERLAPPED Ov;
	Ov.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
	if( Ov.hEvent == NULL )
	{
		LOG_FAILURE_AND_EXIT( L"Failed to allocate event for overlapped operations.", __LINE__ );
	}

	HRESULT hr = NdStartup();
	if( FAILED(hr) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"NdStartup failed with %08x", __LINE__ );
	}

	Timer timer;
	timer.Start();
	if( isServer )
	{
		Server( &Ov, v4Server );
	}
	else
	{
		Client( &Ov, v4Server );
	}
	timer.End();

	printf( "Elapsed time %f seconds\n", timer.Report() / 1000000 );
	hr = NdCleanup();
	if( FAILED(hr) )
	{
		LOG_FAILURE_HRESULT_AND_EXIT( hr, L"NdCleanup failed with %08x", __LINE__ );
	}

	CloseHandle( Ov.hEvent );

	END_LOG( TESTNAME );

	_fcloseall();
	WSACleanup();

	return 0;
}

