// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// ndpingpong.cpp - NetworkDirect send/recv pingpong test
//

#include "ndcommon.h"

const USHORT x_DefaultPort = 54327;
const SIZE_T x_MaxXfer = (4 * 1024 * 1024);
const SIZE_T x_HdrLen = 40;
const SIZE_T x_MaxVolume = (500 * x_MaxXfer);
const SIZE_T x_MaxIterations = 100000;

#define SEND    0
#define BIND    1
#define RECV1   2
#define RECV2   3

void ShowUsage()
{
    printf( "ndrpingpong s|c <ip>[:<port>] [p<nSge>] [l<log>\n"
        "\ts - start as server (listen on IP/Port)\n"
        "\tc - start as client (connect to server IP/Port)\n"
        "\t<ip> - IPv4 Address\n"
        "\t<port> - Port number, default %hu\n"
        "\t<nSge> - Number of scatter/gather entries per transfer, 1 or 2, default 1.\n"
        "\tl - log output to a file named <log>.\n",
        x_DefaultPort
        );
}

void Pong(
    _In_ INDEndpoint* pEndpoint,
    _In_ INDCompletionQueue* pCq,
    _In_ ND_RESULT* Results,
    _In_ ND_SGE* Sgl,
    _In_ const SIZE_T nSge,
    _In_ const ND_MW_DESCRIPTOR* pMwDescriptor,
    _In_ const SIZE_T QueueDepth )
{
    UINT32 *pSize = (UINT32*)Sgl[0].pAddr;
    UINT8 *pBuf = (UINT8*)Sgl[0].pAddr;

    SIZE_T nPosted = 0;
    bool bPong = false;
    HRESULT hr = ND_SUCCESS;
    do
    {
        //
        // Check the CQ for completions - a receive completion indicates
        // that the test is over.
        //
        SIZE_T nResults;
        ND_RESULT* pResult;
        while( (nResults = pCq->GetResults( &pResult, 1 )) != 0 )
        {
            hr = pResult->Status;

            //
            // Receive completions indicate test is over.
            //
            if( pResult == &Results[RECV1] )
            {
                return;
            }

            switch( hr )
            {
            case ND_SUCCESS:
            case ND_CANCELED:
                nPosted--;
                break;

            default:
                printf(
                    "INDCompletionQueue::GetResults returned result with %08x.\n",
                    pResult->Status );
                exit( __LINE__ );
            }

        }// while( nResults != 0 );

        if( !bPong )
        {
            //
            // We wait for the length to be advertised, then go poll on the last
            // byte to detect completion.
            //
            if( *pSize == 0 )
            {
                continue;
            }

            //
            // We now have the size.  Go poll on the last byte.
            //
            if( pBuf[(*pSize) - 1] == 0 )
            {
                continue;
            }

            //
            // Clear the last byte to toggle the sender.
            //
            pBuf[*pSize - 1] = 0;

            //
            // Set the size in our reply area (offset 4).
            //
            *(pSize + 1) = *pSize;

            //
            // Adjust the SGL to pong the size we just received.
            //
            Sgl[nSge - 1].Length = *pSize - (x_HdrLen * (nSge - 1));

            //
            // Clear the size for the next iteration.
            //
            *pSize = 0;

            bPong = true;
        }

        if( bPong && nPosted < QueueDepth )
        {
            //
            // Send the pong.
            //
            hr = pEndpoint->Write( &Results[SEND], Sgl, nSge, pMwDescriptor, 0, 0 );
            if( FAILED( hr ) )
            {
                printf( "INDEndpoint::Write failed with %08x\n", hr );
                exit( __LINE__ );
            }
            nPosted++;
            bPong = false;
        }

    } while( hr == ND_SUCCESS || nPosted != 0 );
}

void Server(
    _In_ OVERLAPPED* pOv,
    _In_ const struct sockaddr_in& v4,
    _In_ SIZE_T nSge,
    _In_ SIZE_T QueueDepth )
{
    INDAdapter* pAdapter;
    HRESULT hr = NdOpenV1Adapter( (const struct sockaddr*)&v4, sizeof(v4), &pAdapter );
    if( FAILED( hr ) )
    {
        printf( "NdOpenV1Adapter failed with %08x\n", hr );
        exit( __LINE__ );
    }

    ND_ADAPTER_INFO Info;
    SIZE_T Len = sizeof(Info);
    hr = pAdapter->Query( 1, &Info, &Len );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::Query failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Allocate and register the data buffer.
    //
    char* pBuf = (char*)HeapAlloc( GetProcessHeap(), 0, x_MaxXfer + x_HdrLen );
    if( !pBuf )
    {
        printf( "Failed to allocate data buffer.\n" );
        exit( __LINE__ );
    }

    ND_MR_HANDLE hMr;
    hr = pAdapter->RegisterMemory( pBuf, x_MaxXfer + x_HdrLen, pOv, &hMr );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::RegisterMemory failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDMemoryWindow* pMw;
    ND_RESULT InvalidateResult;
    hr = pAdapter->CreateMemoryWindow( &InvalidateResult, &pMw );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::CreateMemoryWindow failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Allocate and setup the SGE for all the transfers.
    // Note that all the transfers overlap, so data verification
    // is not possible.
    //
    ND_SGE* Sgl = new (std::nothrow) ND_SGE[nSge];
    if( Sgl == NULL )
    {
        printf( "Failed to allocate SGL.\n" );
        exit( __LINE__ );
    }

    //
    // Listen and get incoming connection request.
    //
    INDListen* pListen;
    hr = pAdapter->Listen( 0, 234, v4.sin_port, NULL, &pListen );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::Listen failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDConnector* pConnector;
    hr = pAdapter->CreateConnector( &pConnector );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::CreateConnector failed with %08x", hr );
        exit( __LINE__ );
    }

    hr = pListen->GetConnectionRequest( pConnector, pOv );
   if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pListen->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDListen::GetConnectionRequest failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Create a CQ.
    //
    INDCompletionQueue* pCq;
    hr = pAdapter->CreateCompletionQueue( (QueueDepth>2?QueueDepth:2) * 2, &pCq );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::CreateCompletionQueue failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDEndpoint* pEndpoint;
    hr = pConnector->CreateEndpoint(
        pCq,
        pCq,
        QueueDepth>2?QueueDepth:2,
        QueueDepth>2?QueueDepth:2,
        nSge,
        nSge,
        0,
        0,
        NULL,
        &pEndpoint
        );
    if( FAILED( hr ) )
    {
        printf( "INDConnector::CreateEndpoint failed with %08x\n", hr );
        exit( __LINE__ );
    }

    ND_RESULT* Results = new (std::nothrow) ND_RESULT[3];
    if( Results == NULL )
    {
        printf( "Failed to allocate ND_RESULT array.\n" );
        exit( __LINE__ );
    }

    for( SIZE_T i = 0; i < nSge; i++ )
    {
        Sgl[i].pAddr = pBuf + (x_HdrLen * i);
        Sgl[i].Length = x_HdrLen;
        Sgl[i].hMr = hMr;
    }

    // Last SGE has the remainder of the data.
    Sgl[nSge - 1].Length = x_MaxXfer + x_HdrLen - (x_HdrLen * (nSge - 1));

    //
    // Pre-post receive request.
    //
    ND_SGE Sge;
    Sge.pAddr = pBuf;
    Sge.Length = sizeof(ND_MW_DESCRIPTOR);
    Sge.hMr = hMr;

    hr = pEndpoint->Receive( &Results[RECV1], &Sge, 1 );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Receive failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Accept the connection, notifying the caller of our queue depth.
    // Note that we send the queue depth in host order (little endian)
    // so that we don't have to worry about 32/64 bit SIZE_T issues.
    // The lower bytes will always be transferred first.
    //
    hr = pConnector->Accept(
        pEndpoint,
        NULL,
        0,
        pOv
        );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pListen->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDListen::Accept failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Exchange MW descriptors.
    //
    SIZE_T nPosted = 1;
    ND_MW_DESCRIPTOR MwDescriptor;
    do
    {
        hr = pCq->Notify( ND_CQ_NOTIFY_ANY, pOv );
        if( hr == ND_PENDING )
        {
            SIZE_T BytesRet;
            hr = pCq->GetOverlappedResult( pOv, &BytesRet, TRUE );
        }
        if( FAILED( hr ) )
        {
            printf( "INDCompletionQueue::Notify failed with %08x\n", hr );
            exit( __LINE__ );
        }

        SIZE_T nResults;
        do
        {
            ND_RESULT* pResult;
            nResults = pCq->GetResults( &pResult, 1 );
            if( nResults == 0 )
            {
                break;
            }

            hr = pResult->Status;

            switch( hr )
            {
            case ND_SUCCESS:
                //
                // Ignore send completions
                //
                if( pResult < &Results[RECV1] )
                {
                    break;
                }

                //
                // If we don't get data, exit.
                //
                if( pResult->BytesTransferred != sizeof(ND_MW_DESCRIPTOR) )
                {
                    printf(
                        "Received bad buffer, expeced %Id bytes, got %Id.\n",
                        sizeof(ND_MW_DESCRIPTOR),
                        pResult->BytesTransferred );
                    exit( __LINE__ );
                }

                //
                // Read the memory window descriptor.
                //
                MwDescriptor = *(ND_MW_DESCRIPTOR*)pBuf;

                //
                // Repost the receive, no data segments.
                // This is to get notification of end of test.
                //
                hr = pEndpoint->Receive( pResult, 0, 0 );
                if( FAILED( hr ) )
                {
                    printf( "INDEndpoint::Receive failed with %08x\n", hr );
                    exit( __LINE__ );
                }

                //
                // Bind the MW.
                //
                hr = pEndpoint->Bind(
                    &Results[BIND],
                    hMr,
                    pMw,
                    pBuf,
                    x_MaxXfer + x_HdrLen,
                    ND_OP_FLAG_ALLOW_READ | ND_OP_FLAG_ALLOW_WRITE,
                    (ND_MW_DESCRIPTOR*)pBuf );
                if( FAILED( hr ) )
                {
                    printf( "INDEndpoint::Bind failed with %08x\n", hr );
                    exit( __LINE__ );
                }

                //
                // Send the MW descriptor.
                //
                hr = pEndpoint->Send( &Results[SEND], &Sge, 1, 0 );
                if( FAILED( hr ) )
                {
                    printf( "INDEndpoint::Send failed with %08x\n", hr );
                    exit( __LINE__ );
                }
                nPosted += 2;
                break;

            case ND_CANCELED:
                printf( "MW Descriptor exchange aborted.\n" );
                exit( __LINE__ );
                break;

            default:
                printf(
                    "INDCompletionQueue::GetResults returned result with %08x.\n",
                    pResult->Status );
                exit( __LINE__ );
            }

            nPosted--;

        } while( nResults != 0 );

    } while( nPosted );

    //
    // Clear the memory buffer.  The first 4 bytes alwasy contain the length
    // of the transfer (including the header), and the last byte is toggled
    // from zero to one to indicate completion of the transfer.
    //
    ZeroMemory( pBuf, x_MaxXfer + x_HdrLen );

    //
    // Send a zero-byte message to indicate we're ready to start.
    //
    hr = pEndpoint->Send( &Results[SEND], NULL, 0, 0 );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Send failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = pCq->Notify( ND_CQ_NOTIFY_ANY, pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDCompletionQueue::Notify failed with %08x\n", hr );
        exit( __LINE__ );
    }

    SIZE_T nResults;
    do
    {
        ND_RESULT* pResult;
        nResults = pCq->GetResults( &pResult, 1 );
        if( nResults == 0 )
        {
            break;
        }

        hr = pResult->Status;

        switch( hr )
        {
        case ND_SUCCESS:
            if( pResult == &Results[RECV1] )
            {
                printf( "Test aborted.\n" );
                exit( __LINE__ );
            }
            break;

        case ND_CANCELED:
            printf( "Test aborted.\n" );
            exit( __LINE__ );
            break;

        default:
            printf(
                "INDCompletionQueue::GetResults returned result with %08x.\n",
                pResult->Status );
            exit( __LINE__ );
        }

    } while( nResults != 0 );

    Pong( pEndpoint, pCq, Results, Sgl, nSge, &MwDescriptor, QueueDepth );

    //
    // Test is complete.  Note that during teardown we don't treat
    // errors as test failures, but we still report them.
    //
    pMw->Release();

    hr = pConnector->Disconnect( pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Disconnect failed wtih %08x\n", hr );
    }

    pEndpoint->Release();
    pConnector->Release();
    pCq->Release();

    pListen->Release();

    hr = pAdapter->DeregisterMemory( hMr, pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::DeregisterMemory failed with %08x\n", hr );
    }

    pAdapter->Release();

    delete[] Results;
    delete[] Sgl;
    HeapFree( GetProcessHeap(), 0, pBuf );
}

HRESULT Ping(
    _In_ INDEndpoint* pEndpoint,
    _In_ INDCompletionQueue* pCq,
    _In_ ND_RESULT* Results,
    _In_ const ND_SGE* Sgl,
    _In_ const SIZE_T nSge,
    _In_ const ND_MW_DESCRIPTOR* pMwDescriptor,
    _In_ SIZE_T QueueDepth,
    _In_ SIZE_T Iterations )
{
    UINT32 *pSize = (UINT32*)Sgl[0].pAddr;
    UINT8 *pBuf = (UINT8*)Sgl[0].pAddr;

    SIZE_T BytesTransferred = 0;
    for( SIZE_T i = 0; i < nSge; i++ )
        BytesTransferred += Sgl[i].Length;

    HRESULT hr = ND_SUCCESS;
    SIZE_T nSendPosted = 0;
    bool NeedSend = true;
    do
    {
        if( Iterations && NeedSend && nSendPosted < QueueDepth )
        {
            //
            // Set the send size.
            //
            *pSize = (UINT32)BytesTransferred;

            //
            // Set the last byte.
            //
            pBuf[BytesTransferred - 1] = 1;

            //
            // Send the ping.
            //
            hr = pEndpoint->Write( &Results[SEND], Sgl, nSge, pMwDescriptor, 0, 0 );
            if( FAILED( hr ) )
            {
                printf( "INDEndpoint::Write failed with %08x\n", hr );
                exit( __LINE__ );
            }
            nSendPosted++;
            NeedSend = false;
        }

        if( nSendPosted > (QueueDepth >> 1) || !Iterations )
        {
            SIZE_T nResults;
            ND_RESULT* pResult;
            while( (nResults = pCq->GetResults( &pResult, 1 )) != 0 )
            {
                hr = pResult->Status;

                switch( hr )
                {
                case ND_SUCCESS:
                    if( pResult->BytesTransferred != BytesTransferred )
                    {
                        printf( "Received %Id bytes, expected %Id\n",
                            pResult->BytesTransferred, BytesTransferred );
                        exit( __LINE__ );
                    }

                    if( pResult >= &Results[RECV1] )
                    {
                        hr = ND_CANCELED;
                        break;
                    }
                    nSendPosted--;

                    __fallthrough;
                case ND_CANCELED:
                    break;

                default:
                    printf(
                        "INDCompletionQueue::GetResults returned result with %08x.\n",
                        pResult->Status );
                    exit( __LINE__ );
                }
            }// while( nResults != 0 )
        }

        if( !NeedSend )
        {
            if( pBuf[BytesTransferred - 1] != 0 )
            {
                continue;
            }

            if( *(pSize + 1) != BytesTransferred )
            {
                printf( "Received %u bytes, expected %Iu\n",
                    *(pSize + 1), BytesTransferred );
                exit( __LINE__ );
            }

            NeedSend = true;
            Iterations--;
        }

    } while( (hr == ND_SUCCESS && Iterations != 0) || nSendPosted > 0 );

    return hr;
}

void Client(
    _In_ OVERLAPPED* pOv,
    _In_ const struct sockaddr_in& v4,
    _In_ SIZE_T nSge,
    _In_ SIZE_T QueueDepth )
{
    struct sockaddr_in LocalAddr;

    SIZE_T Len = sizeof(LocalAddr);
    HRESULT hr = NdResolveAddress(
        (const struct sockaddr*)&v4,
        sizeof(v4),
        (struct sockaddr*)&LocalAddr,
        &Len );
    if( FAILED( hr ) )
    {
        printf( "NdResolveAddress failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDAdapter* pAdapter;
    hr = NdOpenV1Adapter(
        (const struct sockaddr*)&LocalAddr,
        sizeof(LocalAddr),
        &pAdapter );
    if( FAILED( hr ) )
    {
        printf( "NdOpenV1Adapter failed with %08x\n", hr );
        exit( __LINE__ );
    }

    ND_ADAPTER_INFO Info;
    Len = sizeof(Info);
    hr = pAdapter->Query( 1, &Info, &Len );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::Query failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Allocate and register the data buffer.
    //
    char* pBuf = (char*)HeapAlloc( GetProcessHeap(), 0, x_MaxXfer + x_HdrLen );
    if( !pBuf )
    {
        printf( "Failed to allocate data buffer.\n" );
        exit( __LINE__ );
    }

    ND_MR_HANDLE hMr;
    hr = pAdapter->RegisterMemory( pBuf, x_MaxXfer + x_HdrLen, pOv, &hMr );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::RegisterMemory failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDMemoryWindow* pMw;
    ND_RESULT InvalidateResult;
    hr = pAdapter->CreateMemoryWindow( &InvalidateResult, &pMw );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::CreateMemoryWindow failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Allocate and setup the SGE for all the transfers.
    // Note that all the transfers overlap, so data verification
    // is not possible.
    //
    ND_SGE* Sgl = new (std::nothrow) ND_SGE[nSge];
    if( Sgl == NULL )
    {
        printf( "Failed to allocate SGL.\n" );
        exit( __LINE__ );
    }

    //
    // Create a CQ.
    //
    INDCompletionQueue* pCq;
    hr = pAdapter->CreateCompletionQueue( (QueueDepth>2?QueueDepth:2) * 2, &pCq );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::CreateCompletionQueue failed with %08x\n", hr );
        exit( __LINE__ );
    }

    INDConnector* pConnector;
    hr = pAdapter->CreateConnector( &pConnector );
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::CreateConnector failed with %08x", hr );
        exit( __LINE__ );
    }

    INDEndpoint* pEndpoint;
    hr = pConnector->CreateEndpoint(
        pCq,
        pCq,
        QueueDepth>2? QueueDepth : 2,
        QueueDepth>2? QueueDepth : 2,
        nSge,
        nSge,
        0,
        0,
        NULL,
        &pEndpoint );
    if( FAILED( hr ) )
    {
        printf( "INDConnector::CreateEndpoint failed with %08x\n", hr );
        exit( __LINE__ );
    }

    ND_RESULT* Results = new (std::nothrow) ND_RESULT[4];
    if( Results == NULL )
    {
        printf( "Failed to allocate ND_RESULT array.\n" );
        exit( __LINE__ );
    }

    //
    // Connect to the server.
    //
    hr = pConnector->Connect(
        pEndpoint,
        (const struct sockaddr*)&v4,
        sizeof(v4),
        234,
        0,
        NULL,
        0,
        pOv
        );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) || hr == ND_TIMEOUT )
    {
        printf( "INDEndpoint::Connect failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Pre-post receive request.
    //
    ND_SGE Sge;
    Sge.pAddr = pBuf;
    Sge.Length = sizeof(ND_MW_DESCRIPTOR);
    Sge.hMr = hMr;

    hr = pEndpoint->Receive( &Results[RECV1], &Sge, 1 );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Receive failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = pEndpoint->Receive( &Results[RECV2], &Sge, 1 );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Receive failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Complete the connection - this transitions the endpoint so it can send.
    //
    hr = pConnector->CompleteConnect( pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) || hr == ND_TIMEOUT )
    {
        printf( "INDEndpoint::CompleteConnect failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Exchange memory window descriptors.
    //

    //
    // Bind the MW.
    //
    hr = pEndpoint->Bind(
        &Results[BIND],
        hMr,
        pMw,
        pBuf,
        x_MaxXfer + x_HdrLen,
        ND_OP_FLAG_ALLOW_READ | ND_OP_FLAG_ALLOW_WRITE,
        (ND_MW_DESCRIPTOR*)pBuf );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Bind failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Send the MW descriptor.
    //
    hr = pEndpoint->Send( &Results[SEND], &Sge, 1, 0 );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Send failed with %08x\n", hr );
        exit( __LINE__ );
    }
    SIZE_T nPosted = 3;
    SIZE_T nResults = 0;
    ND_MW_DESCRIPTOR MwDescriptor;
    do
    {
        if( nResults == 0 )
        {
            hr = pCq->Notify( ND_CQ_NOTIFY_ANY, pOv );
            if( hr == ND_PENDING )
            {
                SIZE_T BytesRet;
                hr = pCq->GetOverlappedResult( pOv, &BytesRet, TRUE );
            }
            if( FAILED( hr ) )
            {
                printf( "INDCompletionQueue::Notify failed with %08x\n", hr );
                exit( __LINE__ );
            }
        }
        ND_RESULT* pResult;
        nResults = pCq->GetResults( &pResult, 1 );
        if( nResults == 0 )
        {
            continue;
        }

        hr = pResult->Status;

        switch( hr )
        {
        case ND_SUCCESS:
            //
            // Ignore send completions
            //
            if( pResult < &Results[RECV1] )
            {
                break;
            }

            //
            // If we don't get data, exit.
            //
            if( pResult->BytesTransferred != sizeof(ND_MW_DESCRIPTOR) )
            {
                printf(
                    "Received bad buffer, expeced %Id bytes, got %Id.\n",
                    sizeof(ND_MW_DESCRIPTOR),
                    pResult->BytesTransferred );
                exit( __LINE__ );
            }

            //
            // Read the memory window descriptor.
            //
            MwDescriptor = *(ND_MW_DESCRIPTOR*)pBuf;

            //
            // Repost the receive, no data segments.
            // This is to get notification of end of test.
            //
            hr = pEndpoint->Receive( pResult, 0, 0 );
            if( FAILED( hr ) )
            {
                printf( "INDEndpoint::Receive failed with %08x\n", hr );
                exit( __LINE__ );
            }

            __fallthrough;
        case ND_CANCELED:
            break;

        default:
            printf(
                "INDCompletionQueue::GetResults returned result with %08x.\n",
                pResult->Status );
            exit( __LINE__ );
        }

        nPosted--;

    } while( nPosted > 0 );

    printf( "Using %u processors. Sender Frequency is %I64d\n\n"
        " %9s %9s %9s %7s %11s\n",
        CpuMonitor::CpuCount(),
        Timer::Frequency(),
        "Size", "Iter", "Latency", "CPU", "Bytes/Sec"
        );

    //
    // Set all the bytes to 1.  The server will toggle them off.
    //
    FillMemory( pBuf, x_MaxXfer + x_HdrLen, 0x1 );

    //
    // Wait for the server to be ready.
    //
    hr = pCq->Notify( ND_CQ_NOTIFY_ANY, pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pCq->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDCompletionQueue::Notify failed with %08x\n", hr );
        exit( __LINE__ );
    }

    do
    {
        ND_RESULT* pResult;
        nResults = pCq->GetResults( &pResult, 1 );
        if( nResults == 0 )
        {
            break;
        }

        hr = pResult->Status;

        switch( hr )
        {
        case ND_SUCCESS:
            //
            // Ignore send completions
            //
            if( pResult < &Results[RECV1] )
            {
                break;
            }

            //
            // Repost the receive, no data segments.
            // This is to get notification of end of test.
            //
            hr = pEndpoint->Receive( pResult, 0, 0 );
            if( FAILED( hr ) )
            {
                printf( "INDEndpoint::Receive failed with %08x\n", hr );
                exit( __LINE__ );
            }

            __fallthrough;
        case ND_CANCELED:
            break;

        default:
            printf(
                "INDCompletionQueue::GetResults returned result with %08x.\n",
                pResult->Status );
            exit( __LINE__ );
        }

    } while( nResults != 0 );

    //
    // Warm up run, transfer only header.
    //
    Sgl[0].pAddr = pBuf;
    Sgl[0].Length = x_HdrLen;
    Sgl[0].hMr = hMr;

    hr = Ping(
        pEndpoint,
        pCq,
        Results,
        Sgl,
        1,
        &MwDescriptor,
        QueueDepth,
        1000 );
    if( FAILED( hr ) )
    {
        printf( "Connection unexpectedly aborted.\n" );
        exit( __LINE__ );
    }

    Sleep(1000);

    Timer timer;
    CpuMonitor cpu;
    for( SIZE_T szXfer = 1; szXfer <= x_MaxXfer; szXfer <<= 1 )
    {

        for( SIZE_T i = 0; i < nSge; i++ )
        {
            Sgl[i].pAddr = pBuf + (x_HdrLen * i);
            Sgl[i].Length = x_HdrLen;
            Sgl[i].hMr = hMr;
        }

        // Last SGE has the remainder of the data.
        Sgl[nSge - 1].Length = x_HdrLen + szXfer - (x_HdrLen * (nSge - 1));

        SIZE_T Iterations = x_MaxIterations;
        if( Iterations > (x_MaxVolume / szXfer) )
        {
            Iterations = x_MaxVolume / szXfer;
        }

        cpu.Start();
        timer.Start();

        hr = Ping(
            pEndpoint,
            pCq,
            Results,
            Sgl,
            nSge,
            &MwDescriptor,
            QueueDepth,
            Iterations );
        if( FAILED( hr ) )
        {
            printf( "Connection unexpectedly aborted.\n" );
            exit( __LINE__ );
        }

        timer.End();
        cpu.End();

        // Factor of 2 to account for ping *and* pong.
        double bytesSec = 2.0 * szXfer * Iterations / (timer.Report() / 1000000.0);
        // Factor of 2 to account for half-round trip latency.
        double latency = (timer.Report() / Iterations) / 2.0;
        printf(
            " %9Id %9Id %9.2f %7.2f %11.0f\n",
            szXfer,
            Iterations,
            latency,
            cpu.Report(),
            bytesSec
            );
    }

    //
    // Send a message to the server so it exits.
    //
    hr = pEndpoint->Send( &Results[SEND], NULL, 0, 0 );
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Send failed with %08x\n", hr );
        exit( __LINE__ );
    }

    do
    {
        ND_RESULT* pResult;
        nResults = pCq->GetResults( &pResult, 1 );
    } while( nResults == 0 );

    //
    // Test is complete.  Note that during teardown we don't treat
    // errors as test failures, but we still report them.
    //
    pMw->Release();

    hr = pConnector->Disconnect( pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pConnector->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDEndpoint::Disconnect failed wtih %08x\n", hr );
    }

    pEndpoint->Release();
    pConnector->Release();
    pCq->Release();

    hr = pAdapter->DeregisterMemory( hMr, pOv );
    if( hr == ND_PENDING )
    {
        SIZE_T BytesRet;
        hr = pAdapter->GetOverlappedResult( pOv, &BytesRet, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "INDAdapter::DeregisterMemory failed with %08x\n", hr );
    }

    pAdapter->Release();

    delete[] Results;
    delete[] Sgl;
    HeapFree( GetProcessHeap(), 0, pBuf );
}


int __cdecl _tmain(int argc, TCHAR* argv[])
{
    bool bServer = false;
    bool bClient = false;
    SIZE_T nSge = 1;
    SIZE_T QueueDepth = 1;
    struct sockaddr_in v4Server = {0};

    WSADATA wsaData;
    int ret = ::WSAStartup( MAKEWORD( 2, 2 ), &wsaData );
    if( ret != 0 )
    {
        printf( "Failed to initialize Windows Sockets: %d\n", ret );
        exit( __LINE__ );
    }

    for( int i = 1; i < argc; i++ )
    {
        const TCHAR* pArg;
        INT len;

        pArg = argv[i];

        // Skip leading dashes/slashes
        while( *pArg == '-' || *pArg == '/' )
        {
            pArg++;
        }

        switch( *pArg )
        {
        case _T('s'):
        case _T('S'):
            bServer = true;
            if( ++i == argc )
            {
                break;
            }
            len = sizeof( v4Server );
            WSAStringToAddress(
                argv[i],
                AF_INET,
                NULL,
                reinterpret_cast<struct sockaddr*>( &v4Server ),
                &len
                );
            break;
        case _T('c'):
        case _T('C'):
            bClient = true;
            if( ++i == argc )
            {
                break;
            }
            len = sizeof( v4Server );
            WSAStringToAddress(
                argv[i],
                AF_INET,
                NULL,
                reinterpret_cast<struct sockaddr*>( &v4Server ),
                &len
                );
            break;
        case _T('p'):
        case _T('P'):
            nSge = _ttol( ++pArg );
            break;
        case _T('l'):
        case _T('L'):
//
// Can't use freopen_s because it doesn't allow sharing.
// So supress the deprecated warning, because for our use
// it isn't deprecated.
//
#pragma warning( disable : 4996 )
            if( _tfreopen( ++pArg, _T("w"), stdout ) == NULL ||
                _tfreopen( pArg, _T("a+"), stderr ) == NULL )
#pragma warning( default : 4996 )
            {
                printf( "Could not open log file.\n" );
                exit( __LINE__ );
            }
            break;
        default:
            _tprintf( _T("Unknown parameter %s\n"), pArg );
            ShowUsage();
            exit( __LINE__ );
        }
    }

    if( (bClient && bServer) || (!bClient && !bServer) )
    {
        printf( "Exactly one of client (c or "
            "server (s) must be specified.\n" );
        ShowUsage();
        exit( __LINE__ );
    }

    if( v4Server.sin_addr.s_addr == 0 )
    {
        printf( "Bad address.\n\n");
        ShowUsage();
        exit( __LINE__ );
    }

    if( v4Server.sin_port == 0 )
    {
        v4Server.sin_port = htons( x_DefaultPort );
    }

    if( nSge == 0 )
    {
        printf( "Invalid or missing SGE, expected positive integer after p.\n" );
        ShowUsage();
        exit( __LINE__ );
    }

    OVERLAPPED Ov;
    Ov.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
    if( Ov.hEvent == NULL )
    {
        printf( "Failed to allocate event for overlapped operations.\n" );
        exit( __LINE__ );
    }

    HRESULT hr = NdStartup();
    if( FAILED(hr) )
    {
        printf( "NdStartup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    if( bServer )
    {
        Server( &Ov, v4Server, nSge, QueueDepth );
    }
    else
    {
        Client( &Ov, v4Server, nSge, QueueDepth );
    }

    hr = NdCleanup();
    if( FAILED(hr) )
    {
        printf( "NdCleanup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    CloseHandle( Ov.hEvent );

    _fcloseall();
    WSACleanup();

    return 0;
}
